#include "map.h"
#include "math.h"
#include "checksum.h"


//////// Tile data ////////
// Assembly rips by Snrrrub and Coconut Emulator

Uint32 makeTileData(Uint16 x, Uint16 y, BYTE type)
{
	return (x & 0xFFF) | ((y & 0xFFF) << 12) | (type << 24);
}

tileData makeTileData(Uint32 raw)
{
	tileData t;

	t.x = Uint16(raw & 0xFFF);
	t.y = Uint16((raw >> 12) & 0xFFF);
	t.type = BYTE(raw >> 24);

	return t;
}


//////// Map coordinates ////////
// Enormous amount of timed saved by FACTS team

Uint32 getLinear(Uint32 x, Uint32 y)
{
	return (y << 10) + x;
}

void convertFileToMatrix(char * fileData, char * mapData, Uint32 len)
{
	memset(mapData, 0, 0x100000);

	// Recurse to skip a tileset (if present)
	if ( (fileData[0] == 'B') &&
		 (fileData[1] == 'M')  )
	{
		Uint32 diff = getLong(fileData, 2);
		convertFileToMatrix(fileData + diff, mapData, len - diff);

		return;
	}

	// Fill map-tiles
	for (Uint32 i = 0; i < len; i += 4)
	{
		tileData t = makeTileData(getLong(fileData, i));

		mapData[getLinear(t.x, t.y)] = t.type;
	}

	// Fill border-tiles
	mapData[0] = vieBorder;
	mapData[getLinear(0, 1023)] = vieBorder;
	mapData[getLinear(1023, 0)] = vieBorder;
	mapData[getLinear(1023, 1023)] = vieBorder;
	for (Uint16 x = 1; x < 1023; x++)
	{
		mapData[getLinear(x,	0)] = vieBorder;
		mapData[getLinear(x, 1023)] = vieBorder;
		mapData[getLinear(0,	x)] = vieBorder;
		mapData[getLinear(1023,	x)] = vieBorder;
	}
}

Uint32 getMapSize(char * mapData)
{
	Uint32 offset = 0;

	for (Uint16 x = 1; x < 1023; x++)		// Ignore border
	for (Uint16 y = 1; y < 1023; y++)
	{
		BYTE t = mapData[getLinear(x, y)];

		if (t == 0)
			continue;

		if ( ((t >= vieNormalStart) && (t <= vieFlyUnderEnd)) ||	// Regular tiles
			 ((t >= vieAsteroidStart) && (t <= vieWormhole)) )		// Animated tiles
		{	// The tile is not internal
			offset++;
		}
	}

	return (offset << 2);
}

void convertMatrixToFile(char * mapData, char * _fileData)
{
	Uint32 offset = 0;
	Uint32 * fileData = (Uint32*)_fileData;

	for (Uint16 x = 1; x < 1023; x++)		// Ignore border
	for (Uint16 y = 1; y < 1023; y++)
	{
		BYTE t = mapData[getLinear(x, y)];

		if (t == 0)
			continue;

		if ( ((t >= vieNormalStart) && (t <= vieFlyUnderEnd)) ||	// Regular tiles
			 ((t >= vieAsteroidStart) && (t <= vieWormhole)) )		// Animated tiles
		{	// The tile is not internal
			fileData[offset++] = makeTileData(x, y, t);
		}
	}
}


//////// Tile synchronization ////////

prizeSystem::prizeSystem()
{
	rlevel = 0;
}

Uint32 prizeSystem::getTime()
{
	return GetTickCount() / 10;
}

Uint32 prizeSystem::getNumberOfGreens(Uint16 population)
{
	long total = IMULHIDWORD(prizeFactor * population, cosmicPrize);
	total = (total >> 6) + (((Uint32)total) >> 31);
	total = limit(total, 256);

	long created = total - list.total;

	return (created > 0) ? created : 0;
}

void prizeSystem::decayGreens()
{
	_listnode <prize> * parse = list.head;

	while (parse)
	{
		prize * p = parse->item;
		parse = parse->next;

		if (getTime() > p->time)
		{
			map[getLinear(p->x, p->y)] = vieNoTile;
			list.kill(p, false);
		}
	}
}

prize * prizeSystem::createGreen(Uint16 population)
{
	prize * p = new prize;

	Uint32 stop;
	Uint32 current;

	// Find dimensions of the prize bounding rectangle
	Uint16 distance = population * prizeUpgradeVirtual + prizeMinimumVirtual;
	distance = (Uint16)limit(distance, 1024);

	// Fill prize coordinates
	p->x = (prng.getNextG() % (distance - 2)) + ((1024 - distance) >> 1) + 1;
	p->y = (prng.getNextG() % (distance - 2)) + ((1024 - distance) >> 1) + 1;

	// Fill prize type
	p->type = 0;
	current = 0;
	stop	= prng.getNextG() % prizeWeightTotal;
	BYTE * pw = (BYTE*)&prizeWeights;
	for (int i = 0; i < 28; i++)
	{
		if (current < stop) break;

		current += (*pw++);
		p->type++;
	}

	// Determine if prize type is negative or not
	if (prng.getNextG() % prizeNegativeFactor == 0)
		p->type = -p->type;

	// Fill alive time
	long time = prizeMaxExist - prizeMinExist;
	Uint16 factor = prng.getNextG();
	if (time) time = factor % time;
	p->time = time + prizeMinExist + getTime();

	return p;
}

void prizeSystem::setSettings(arenaSettings * s, BYTE * m)
{
	map = m;

	// Fill settings cache
	prizeFactor = s->PrizeFactor;
	prizeDelay = s->PrizeDelay;
	prizeHideCount = s->PrizeHideCount;
	prizeUpgradeVirtual = s->PrizeUpgradeVirtual;
	prizeMinimumVirtual = s->PrizeMinimumVirtual;
	prizeMaxExist = s->PrizeMaxExist;
	prizeMinExist = s->PrizeMinExist;
	prizeNegativeFactor = s->PrizeNegativeFactor;
	MEMCPY(&s->pw, &prizeWeights, sizeof(prizeSettings));

	// Compute prize weight total
	prizeWeightTotal = 0;
	BYTE * pw = (BYTE*)&prizeWeights;
	for (int i = 0; i < 28; i++)
	{
		prizeWeightTotal += (*pw++);
	}

	rlevel++;
}

void prizeSystem::randomize(Uint32 seed, Uint32 time)
{
	prng.seed(seed);

	rlevel++;
	lastTime = time;
}

void prizeSystem::doPrizes(Uint16 population)
{
	if (rlevel < 2)
	{
		return;
	}

	long iterations = (getTime() - lastTime) / prizeDelay;

	if (iterations == 0)
		return;

	lastTime = getTime();

	decayGreens();

	Uint32 count = getNumberOfGreens(population);

	for (int i = 0; i < iterations; i++)
	{
		for (int j = 0; j < prizeHideCount; j++)
		{
			prize * p = createGreen(population);

			if (count)
			{
				Uint32 offset = getLinear(p->x, p->y);

				if (map[offset])
				{
					delete p;
				}
				else
				{
					map[offset] = ssbPrize;
					count--;

					list.append(p);
				}
			}
			else
			{
				delete p;
			}
		}
	}
}

void prizeSystem::clearList()
{
	list.clear(false);
}
